#include "GLWindow.h"
#include <iostream>
#include "ngl/Camera.h"
#include "ngl/Transformation.h"
#include "ngl/TransformStack.h"
#include "ngl/NGLInit.h"
#include "ngl/ShaderManager.h"
#include "ngl/Util.h"
#include "Noise.h"
#include <QImage>
#include <ctime>

/// On the mac these are extentions, it's easier to let the
/// pre-processor deal with this than using GLEW / other code
/// as on Win and Linux it works as normal glGen

#ifdef DARWIN
  #define glGenVertexArrays glGenVertexArraysAPPLE
  #define glBindVertexArray glBindVertexArrayAPPLE
#endif


//----------------------------------------------------------------------------------------------------------------------
GLWindow::GLWindow(
                   QWidget *_parent
                  ) :
                    QGLWidget(_parent)
{

  // set this widget to have the initial keyboard focus
  setFocus();
  // re-size the widget to that of the parent (in this case the GLFrame passed in on construction)
  this->resize(_parent->size());
  // Now set the initial GLWindow attributes to default values
  // Roate is false
  m_rotate=false;
  // mouse rotation values set to 0
  m_spinXFace=0;
  m_spinYFace=0;
  m_origX=0;
  m_origY=0;
	m_fpsTimer =startTimer(0);
	  m_fps=0;
	  m_frames=0;
	  m_timer.start();
}

GLWindow::~GLWindow()
{
  ngl::NGLInit *Init = ngl::NGLInit::instance();
  std::cout<<"Shutting down NGL, removing VBO's and Shaders\n";
  Init->NGLQuit();
}


void GLWindow::makeMarbleTexture(float amp, float strength)
{
// create a new instance of the noise class (which also creates the lattice noise tables)
Noise *n = new Noise();
// pointer to the Texture data
float *data;
// size of the texture width
const static int MSIZE=256;
// make room for the texture (only using Luminance value here)
data = new float [MSIZE*MSIZE*MSIZE];
// calc step size for S and T rangeing from 0.0 1.0
float step=1.0/(float)MSIZE;
float S=0.0;
float T=0.0;
float U=0.0;
// create index into data set
unsigned int index=0;
// loop for width and height for the texture to create
std::cout <<"Creating texture"<<std::endl;
ngl::Vector p;
for(int z=0; z<MSIZE; ++z)
	{
	for(int y=0; y<MSIZE; ++y)
	{
	for(int x=0; x<MSIZE; ++x)
		{
		// the marble functions requires an input of a point in 3d space
		// we use S and T for the x,y and vary Z along the U axis
		p.set(S,T,U);
		// now get the value from the marble function (from noise class)
		// amp and strength are global variables set at the start of the program
		// or from the command line
		data[index]=n->marble(amp,strength,p);
		// inc to next data set
		++index;
		S+=step;
		}
	// inc T and reset S
	T+=step;
	S=0.0;
	}
	T=0.0;
	U+=step;
	}
// now assign the texture object
GLuint textureName;
glGenTextures(1,&textureName);
glBindTexture(GL_TEXTURE_3D,textureName);
glTexParameteri(GL_TEXTURE_3D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
glTexParameteri(GL_TEXTURE_3D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
glTexParameteri(GL_TEXTURE_3D,GL_TEXTURE_WRAP_S,GL_REPEAT);
glTexParameteri(GL_TEXTURE_3D,GL_TEXTURE_WRAP_T,GL_REPEAT);
glTexParameteri(GL_TEXTURE_3D,GL_TEXTURE_WRAP_R,GL_REPEAT);
glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND);



glTexImage3D(GL_TEXTURE_3D,0,GL_LUMINANCE,MSIZE,MSIZE,MSIZE,0,GL_LUMINANCE,GL_FLOAT,data);
std::cout <<"done texture"<<std::endl;
// remove the Data
delete [] data;
// return the id of the texture object
}





// This virtual function is called once before the first call to paintGL() or resizeGL(),
//and then once whenever the widget has been assigned a new QGLContext.
// This function should set up any required OpenGL context rendering flags, defining display lists, etc.

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::initializeGL()
{

  glClearColor(0.4f, 0.4f, 0.4f, 1.0f);			   // Grey Background
  // enable depth testing for drawing
   glEnable(GL_DEPTH_TEST);
   // Now we will create a basic Camera from the graphics library
   // This is a static camera so it only needs to be set once
   // First create Values for the camera position
   ngl::Vector From(1,1,1);
   ngl::Vector To(0,0,0);
   ngl::Vector Up(0,1,0);
   ngl::NGLInit *Init = ngl::NGLInit::instance();
   #ifdef WIN32
   	glewInit();
   #endif
   Init->initGlew();
   m_cam= new ngl::Camera(From,To,Up,ngl::PERSPECTIVE);
   // set the shape using FOV 45 Aspect Ratio based on Width and Height
   // The final two are near and far clipping planes of 0.5 and 10
   m_cam->setShape(45,(float)720.0/576.0,0.5,150,ngl::PERSPECTIVE);
   // now to load the shader and set the values
   // grab an instance of shader manager
   ngl::ShaderManager *shader=ngl::ShaderManager::instance();
   // load a frag and vert shaders

	shader->createShaderProgram("TextureShader");

	shader->attachShader("SimpleVertex",ngl::VERTEX);
	shader->attachShader("SimpleFragment",ngl::FRAGMENT);
	shader->loadShaderSource("SimpleVertex","shaders/Vertex.vs");
	shader->loadShaderSource("SimpleFragment","shaders/Fragment.fs");

	shader->compileShader("SimpleVertex");
	shader->compileShader("SimpleFragment");
	shader->attachShaderToProgram("TextureShader","SimpleVertex");
	shader->attachShaderToProgram("TextureShader","SimpleFragment");


	shader->linkProgramObject("TextureShader");
	shader->use("TextureShader");

   // now pass the modelView and projection values to the shader
   shader->setShaderParamFromMatrix("TextureShader","ViewMatrix",m_cam->getModelView());
   shader->setShaderParamFromMatrix("TextureShader","projectionMatrix",m_cam->getProjection());

   glEnable(GL_DEPTH_TEST); // for removal of hidden surfaces
	makeMarbleTexture(0.00007f,18.0);
}

//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget has been resized.
// The new size is passed in width and height.
void GLWindow::resizeGL(
                        int _w,
                        int _h
                       )
{
  glViewport(0,0,_w,_h);
  m_cam->setShape(45,(float)_w/_h,0.5,150,ngl::PERSPECTIVE);
  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
  shader->setShaderParamFromMatrix("TextureShader","projectionMatrix",m_cam->getProjection());
}

//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget needs to be painted.
// this is our main drawing routine
void GLWindow::paintGL()
{
  // clear the screen and depth buffer
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
 // build our transform stack
  ngl::Transformation trans;
  // set the mouse rotation
  trans.setRotation(m_spinXFace,m_spinYFace,0);
  // set this in the TX stack
  m_transformStack.setGlobal(trans);

  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
	(*shader)["TextureShader"]->use();
  m_transformStack.loadGlobalAndCurrentMatrixToShader("TextureShader","ModelMatrix");

	shader->setShaderParamFromMatrix("TextureShader","ModelMatrix",m_transformStack.getCurrAndGlobal().getMatrix());
	// now we bind back our vertex array object and draw
	ngl::VBOPrimitives *prim =  ngl::VBOPrimitives::instance();
	prim->draw("teapot");
}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mouseMoveEvent (
                               QMouseEvent * _event
                              )
{
  // note the method buttons() is the button state when event was called
  // this is different from button() which is used to check which button was
  // pressed when the mousePress/Release event is generated
  if(m_rotate && _event->buttons() == Qt::LeftButton)
  {
    m_spinYFace = ( m_spinYFace + (_event->x() - m_origX) ) % 360 ;
    m_spinXFace = ( m_spinXFace + (_event->y() - m_origY) ) % 360 ;
    m_origX = _event->x();
    m_origY = _event->y();
  }
  // re-draw GL
  updateGL();
}


//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mousePressEvent (
                                QMouseEvent * _event
                               )
{
  // this method is called when the mouse button is pressed in this case we
  // store the value where the maouse was clicked (x,y) and set the Rotate flag to true
  if(_event->button() == Qt::LeftButton)
  {
    m_origX = _event->x();
    m_origY = _event->y();
    m_rotate =true;
  }
}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mouseReleaseEvent (
                                  QMouseEvent * _event
                                 )
{
  // this event is called when the mouse button is released
  // we then set Rotate to false
  if (_event->button() == Qt::LeftButton)
  {
    m_rotate=false;
  }
}

void GLWindow::timerEvent(
                          QTimerEvent *_event
                         )
{
	if(_event->timerId() == m_fpsTimer)
	  {
	    if( m_timer.elapsed() > 1000.0)
	    {
	      m_fps=m_frames;
	      m_frames=0;
	      m_timer.restart();
	    }
	   }
	    // re-draw GL
	updateGL();
}

